import rumd
import time


def ResetMomentum(sample):
    # Made a global function in order to incorporate into callback mechanism
    # Defining it as a method of rumdSimulation somehow prevents the
    # rumdSimulation object being destroyed when it's supposed to be.
    # Must be some kind of circular reference (self.callbackList then contains
    # a method object which involves the instance object )
    sample.GetIntegrator().SetMomentumToZero()
    
class rumdSimulation:

    """ Contains the main objects and provides basic simulation main loop"""
    
    def __init__(self, filename, pb=32, tp=4, verbose=True):
        self.sample = rumd.Sample(pb,tp)
        self.sample.SetVerbose(verbose)
        self.sample.ReadConf(filename)
               
        self.thermostat = False
        
        self.outputParameters = {}
        self.outputParameters["energies"] = ("linear", 256)
        self.outputParameters["trajectory"] =  ("logarithmic",1)

        self.itg = None
        self.potentialList = []
        self.output_is_initialized = False
        self.blockSize = None
        self.moleculeData = None
        self.verbose = verbose
        self.callbackList = [(ResetMomentum, 100)]

    def WriteConstraints(self):
        self.moleculeData.WriteConstraints()

    def SetVerbose(self, vb):
        """Turn on/off most messages in rumdSimulation and sample objects"""
        self.verbose = vb
        self.sample.SetVerbose(self.verbose)

    def ReadMoleculeData(self, topologyFile):
        """Read data describing how particles are connected as molecules"""
        self.moleculeData = rumd.MoleculeData(self.sample, topologyFile)
        
    def SetBondHarmonic(self, type, lbond, ks, exclude=True):
        """Set a bond-type to be harmonic"""
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        self.moleculeData.SetBondHarmonic(type, lbond, ks)

        if exclude:
            if len(self.potentialList) == 0:
                print "Warning: SetBondHarmonic called before non-bonding potential has been set. Cannot set exclusion list"
            else:
                for potential in self.potentialList:
                    self.moleculeData.SetExclusionBond(potential, type)
            
          
    def SetBondFENE(self, type):
        """Set a bond-type to be the FENE type"""
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        self.moleculeData.SetBondFENE(type)
        
    def SetBondConstraint(self, type, lbond, exclude=True):
        """Set a bond-type as "constraint" and pass the length"""
        if self.moleculeData is None:
            raise ValueError, "No moleculeData object; call ReadMoleculeData first"
        
        self.moleculeData.SetBondConstraint(type, lbond)
        if exclude:
            if len(self.potentialList) == 0:
                print "Warning: SetBondConstraint called before non-bonding potential has been set. Cannot set exclusion list"
            else:
                for potential in self.potentialList:
                    self.moleculeData.SetExclusionBond(potential, type)
                    

    def SetPotential(self, potential):
        """Set a potential for this system"""

        self.potentialList = [potential]
        self.sample.SetPotential(potential)

    def AddPotential(self, potential):
        """Add a potential for this system"""

        self.potentialList.append(potential)
        self.sample.AddPotential(potential)

    def SetIntegrator(self, itg):
        """Set the integrator for this system"""
        self.sample.SetIntegrator(itg)
        self.itg = itg
        itg.SetMomentumToZero()
        
    def WriteConf(self, filename):
        """Write the current configuration to a file"""
        self.sample.WriteConf(filename)

    def SetBlockSize(self, blockSize):
        """Specify the block-size for output (None means automatic)"""
        self.blockSize = blockSize
        
    def SetOutputParameters(self, manager, schedule, **kwargs):
        """Set scheduling information for an output manager """
        allowedManagers = ["energies","trajectory"]
        allowedSchedules = ["none", "linear","logarithmic","loglin"]

        if manager not in allowedManagers:
            raise ValueError, "SetOutputParameters -- first argument must be one of" + str(allowedManagers)
        if schedule not in allowedSchedules:
            raise ValueError, "SetOutputParameters -- second argument must be " + str(allowedSchedules)

        if schedule == "none":
            newParams = None
        elif schedule == "linear":
            newParams = ("linear",kwargs["interval"])
        elif schedule == "logarithmic":
            base = 1
            if kwargs.has_key("base"):
                base = kwargs["base"]
            newParams = ("logarithmic",base)
        else:
            newParams = ("loglin",kwargs["base"],kwargs["maxInterval"])

        self.outputParameters[manager] = newParams


    def InitializeOutput(self, num_iter):
        """Not intended to be called by the user"""
        SetLinearFuncMap = {"energies":self.sample.SetEnergyLinearWriteInterval,
                         "trajectory":self.sample.SetTrajectoryLinearWriteInterval}
        SetLogLinFuncMap = {"energies":self.sample.SetEnergyLogLinParams,
                            "trajectory":self.sample.SetTrajectoryLogLinParams}

        SetOnFuncMap = {"energies":self.sample.SetWriteEnergies,
                        "trajectory":self.sample.SetWriteTrajectory}

        ON = {}
        min_blockSize = 1;
        for manager in self.outputParameters.keys():
            outputParameters = self.outputParameters[manager]
            ON[manager] = (outputParameters is not None)
            SetOnFuncMap[manager](ON[manager])
            if ON[manager]:
                schedule = outputParameters[0]
                if schedule == "linear":
                    interval = outputParameters[1]
                    SetLinearFuncMap[manager](interval)
                    if interval > min_blockSize: min_blockSize = interval
                elif schedule == "logarithmic":
                    base = outputParameters[1]
                    SetLogLinFuncMap[manager](base, 0)
                else:
                    base = outputParameters[1]
                    maxInterval = outputParameters[2]
                    SetLogLinFuncMap[manager](base, maxInterval)

        if self.blockSize is None:
            nominalNumBlocks = 100
            if (ON["energies"] or ON["trajectory"]):
                nominalBlockSize = num_iter/nominalNumBlocks/2
                if min_blockSize < nominalBlockSize:
                    min_blockSize = nominalBlockSize
                blockSize = 1
                while blockSize < min_blockSize:
                    blockSize *= 2
            else:
                blockSize = num_iter/nominalNumBlocks
            
            self.blockSize = blockSize
        numBlocks = num_iter/self.blockSize + 1

        if self.verbose:
            print "Output blockSize=%d, numBlocks=%d" % (self.blockSize, numBlocks)
            print "Size of the last incomplete block", num_iter % self.blockSize
            

        self.sample.SetOutputBlockSize(self.blockSize)
        self.output_is_initialized = True


    def RegisterCallback(self, function, interval):
        """ Register a data analysis function (which takes a Sample as argument) to be called with a specified time-step interval """
        self.callbackList.append((function, interval))

    def SetMomentumResetInterval(self, mom_reset_interval):
        """ Set how many time steps should go between resetting of
        center of mass momentum to zero"""
        self.callbackList[0] = (ResetMomentum, mom_reset_interval)
        
    def Run(self, num_iter, initializeOutput=True, restartBlock=None, suppressAllOutput=False):
        """Run a given number of time steps"""

        if self.itg is None:
            raise ValueError, "Integrator has not been set"
        if self.potentialList is None:
            raise ValueError, "Potential has not been set"

        initialTimeStep = 0
        if not suppressAllOutput:
            if initializeOutput:
                self.InitializeOutput(num_iter)
            else:
                if not self.output_is_initialized:
                    raise ValueError, "Cannot have initializeOutput=False if output has not previously been initialized"

            if restartBlock is not None:
                if not initializeOutput:
                    raise ValueError, "When restarting, initializeOutput must be True"
                self.sample.ReadRestartConf(restartBlock)
                initialTimeStep = restartBlock*self.blockSize


        for i in xrange(initialTimeStep, num_iter):
            self.sample.CalcF() # Calculate forces
            if not suppressAllOutput:
                self.sample.NotifyOutputManagers(i) # Called here to ensure synchronization of positions and forces

            # user supplied data processing functions (call-backs)
            for (func, interval) in self.callbackList:
                if interval > 0 and i % interval == 0 and i > 0:
                    func(self.sample) 

            self.itg.Integrate() # Update positions and velocities
 

        # end main loop

