
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/Sample.h"
#include "rumd/rumd_algorithms.h"
#include "rumd/RUMD_Error.h"
#include "rumd/IntegratorNVU.h"
#include "rumd/Potential.h"

const std::string NVU_Error_Code1("This integrator is not associated with a Sample or the latter has no particles");
const std::string NVU_Error_Code2("Failed to dynamic_cast simulation box to RectangularSimulationBox");

////////////////////////////////////////////////////////////
// Constructors 
////////////////////////////////////////////////////////////

IntegratorNVU::IntegratorNVU() {
  timeStep = 1.f;
  displacementLength = 0.1f;
  targetPotentialEnergy = -1.f;
  numParticlesSample = 0;
  h_previousPotentialEnergy = (float*) malloc( sizeof(float) );

  if( cudaMalloc( (void**) &d_previousPotentialEnergy, sizeof(float) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("IntegratorNVU","IntegratorNVU","Malloc failed on d_previousPotentialEnergy") );
}

IntegratorNVU::~IntegratorNVU(){
  free(h_previousPotentialEnergy);
  cudaFree(d_previousPotentialEnergy);
  FreeArrays();
}

////////////////////////////////////////////////////////////
// Class Methods
////////////////////////////////////////////////////////////

void IntegratorNVU::Integrate(){
  if(!S || !(S->GetNumberOfParticles()) || numParticlesSample != S->GetNumberOfParticles())
    throw( RUMD_Error("IntegratorNVU","Integrate", NVU_Error_Code1 ) );
  
  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(S->GetSimulationBox());
  if(!testRSB) throw RUMD_Error("IntegratorNVU","Integrate", NVU_Error_Code2);
  
  integrateNVUAlgorithm<<< kp.grid, kp.threads >>>( numParticlesSample, P->d_r, P->d_v, P->d_f, 
						    P->d_im, testRSB, testRSB->GetDevicePointer(), d_summationArrayForce, 
						    d_summationArrayCorrection, d_previousPotentialEnergy, 
						    numParticlesSample * targetPotentialEnergy, displacementLength );
  
  float* forwardP = &d_summationArrayCorrection[numParticlesSample];
  cudaMemcpy( d_previousPotentialEnergy, forwardP, sizeof(float), cudaMemcpyDeviceToDevice );
}

// Constrains the motion on U.
void IntegratorNVU::CalculateConstraintForce() {
  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(S->GetSimulationBox());
  if(!testRSB) throw RUMD_Error("IntegratorNVU","Integrate", NVU_Error_Code2);
  
  calculateConstraintForce<<< kp.grid, kp.threads >>>( numParticlesSample, P->d_r, d_euclideanConfigurationPos, 
						       P->d_im, d_euclideanConfigurationImg,
						       P->d_v, P->d_f, d_summationArrayForce, testRSB, 
						       testRSB->GetDevicePointer() );

  sumIdenticalArrays( d_summationArrayForce, numParticlesSample, 3, 32 );
}

// Private. Step length and potential energy correction.
void IntegratorNVU::CalculateNumericalCorrection(){
  calculateNumericalCorrection<<< kp.grid, kp.threads >>>( numParticlesSample, P->d_v, P->d_f, d_summationArrayForce, 
							   d_summationArrayCorrection, d_previousPotentialEnergy, 
							   numParticlesSample * targetPotentialEnergy );
  
  sumIdenticalArrays( d_summationArrayCorrection, numParticlesSample, 2, 32 );
}

void IntegratorNVU::FreeArrays(){
  if(numParticlesSample){
    cudaFree(d_euclideanConfigurationPos);
    cudaFree(d_euclideanConfigurationImg);
    cudaFree(d_particleVelocity);
    cudaFree(d_summationArrayForce);
    cudaFree(d_summationArrayCorrection);
  }
}

void IntegratorNVU::AllocateIntegratorState(){
  if(!S || !(S->GetNumberOfParticles())) // Consistency check with itg.
    throw( RUMD_Error("IntegratorNVU","AllocateIntegratorState", NVU_Error_Code1 ) );
  
  unsigned int newNumParticlesS = S->GetNumberOfParticles();
  
  if(newNumParticlesS != numParticlesSample){
    FreeArrays();
    numParticlesSample = newNumParticlesS;
    
    h_previousPotentialEnergy[0] = numParticlesSample * targetPotentialEnergy;
    cudaMemcpy( d_previousPotentialEnergy, h_previousPotentialEnergy, sizeof(float), cudaMemcpyHostToDevice );
    
    // Blocks until the device has completed all preceding requested tasks. Host => Device is async.
    if( cudaThreadSynchronize() != cudaSuccess ) 
      throw( RUMD_Error("IntegratorNVU","AllocateIntegratorState","CudaMemcpy failed: h_previousPotentialEnergy => d_previousPotentialEnergy") );
    
    // Allocate space on the GPU for the summation of the constraint force.
    if( cudaMalloc( (void**) &d_summationArrayForce, 3 * numParticlesSample * sizeof(float) ) == cudaErrorMemoryAllocation ) 
      throw( RUMD_Error("IntegratorNVU","AllocateIntegratorState","Malloc failed on d_summationArrayForce") );
    
    // Allocate space on the GPU for the summation of the numerical correction.
    if( cudaMalloc( (void**) &d_summationArrayCorrection, 2 * numParticlesSample * sizeof(float) ) == cudaErrorMemoryAllocation ) 
      throw( RUMD_Error("IntegratorNVU","AllocateIntegratorState","Malloc failed on d_summationArrayCorrection") );
    
    // Allocate space for the summation of the system velocity on GPU.
    if( cudaMalloc( (void**) &d_particleVelocity, numParticlesSample * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
      throw( RUMD_Error("IntegratorNVU","AllocateIntegratorState","Malloc failed on d_particleVelocity") );
    
    if( cudaMalloc( (void**) &d_euclideanConfigurationPos, numParticlesSample * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
      throw( RUMD_Error("IntegratorNVU","AllocateIntegratorState","Malloc failed on d_euclideanConfigurationPos") );
    
    if( cudaMalloc( (void**) &d_euclideanConfigurationImg, numParticlesSample * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
      throw( RUMD_Error("IntegratorNVU","AllocateIntegratorState","Malloc failed on d_euclideanConfigurationImg") );
    
    // Currently disabled in NVU.
    S->GetEnergiesMetaData().Set("totalEnergy",false);
    S->GetEnergiesMetaData().Set("temperature",false);
    S->GetEnergiesMetaData().Set("kineticEnergy",false);
    S->GetEnergiesMetaData().Set("pressure", false);
    S->GetEnergiesMetaData().Set("thermostat_Ps", false);
  }
}

void IntegratorNVU::CalculateAfterForce(){
  if(!S || !(S->GetNumberOfParticles()) || numParticlesSample != S->GetNumberOfParticles())
    throw( RUMD_Error("IntegratorNVU","CalculateConstraintForce", NVU_Error_Code1 ));
  
  CalculateConstraintForce();
  CalculateNumericalCorrection(); 
}

////////////////////////////////////////////////////////////
// Set Methods
////////////////////////////////////////////////////////////

void IntegratorNVU::SetDisplacementLength( float step ){ displacementLength = step; }
void IntegratorNVU::SetTargetPotentialEnergy( float U ){ targetPotentialEnergy = U; };

void IntegratorNVU::SetMomentumToZero(){
  unsigned int nParticlesS;
  if(!S || !(nParticlesS = S->GetNumberOfParticles()))
    throw( RUMD_Error("IntegratorNVU","SetMomentumToZero", NVU_Error_Code1 ) );
  
  calculateVelocity<<< kp.grid, kp.threads >>>( nParticlesS, P->d_v, d_particleVelocity );
  sumIdenticalArrays( d_particleVelocity, numParticlesSample, 1, 32 );
  zeroSystemVelocity<<<kp.grid, kp.threads>>>( nParticlesS, P->d_v, d_particleVelocity );
}

void IntegratorNVU::SetEuclideanConfiguration(){
  unsigned int nParticlesS;
  if(!S || !(nParticlesS = S->GetNumberOfParticles()))
    throw( RUMD_Error("IntegratorNVU","SetEuclideanConfiguration", NVU_Error_Code1 ) );

  cudaMemcpy( d_euclideanConfigurationPos, P->d_r, nParticlesS * sizeof(float4), cudaMemcpyDeviceToDevice );
  cudaMemcpy( d_euclideanConfigurationImg, P->d_im, nParticlesS * sizeof(float4), cudaMemcpyDeviceToDevice );
}

////////////////////////////////////////////////////////////
// Get Methods 
////////////////////////////////////////////////////////////

float IntegratorNVU::GetSimulationTimeStepSq() const { 
  if(!S || !(S->GetNumberOfParticles()))
    throw( RUMD_Error("IntegratorNVU","GetSimulationTimeStepSq", NVU_Error_Code1 ) );
  
  float* forwardP = &d_summationArrayForce[numParticlesSample];
  float constraintForceNumerator = 0; float constraintForceDenominator = 1.f;
  cudaMemcpy( &constraintForceNumerator, d_summationArrayForce, sizeof(float), cudaMemcpyDeviceToHost );
  cudaMemcpy( &constraintForceDenominator, forwardP, sizeof(float), cudaMemcpyDeviceToHost );
  return constraintForceNumerator/constraintForceDenominator; 
}

float IntegratorNVU::GetDisplacementLength() const { return displacementLength; }

float IntegratorNVU::GetSimulationDisplacementLength() const { 
  unsigned int nParticlesS = S->GetNumberOfParticles();
  if(!nParticlesS)
    throw( RUMD_Error("IntegratorNVU","GetSimulationDisplacementLength","Sample has no particles" ) );
  
  P->CopyVelFromDevice();
  float dispLength = 0;
  for(unsigned int i=0; i < S->GetNumberOfParticles(); i++) 
    dispLength += P->h_v[i].x * P->h_v[i].x + P->h_v[i].y * P->h_v[i].y + P->h_v[i].z * P->h_v[i].z;
  return sqrtf(dispLength);
} 

float IntegratorNVU::GetTargetPotentialEnergy() const { return targetPotentialEnergy; };

float IntegratorNVU::GetEuclideanLength() const { 
  unsigned int nParticlesS = S->GetNumberOfParticles();
  if(!S->GetIntegrator() || !nParticlesS)
    throw( RUMD_Error("IntegratorNVU","GetEuclideanLength", NVU_Error_Code2 ) );

  float* forwardP = &d_summationArrayForce[2*numParticlesSample];
  float euclideanLength = 0;
  cudaMemcpy( &euclideanLength, forwardP, sizeof(float), cudaMemcpyDeviceToHost );
  return sqrtf(euclideanLength);
} 

//////////////////////////////////////////////////////////////////////////////////////////////////////
// NVU algorithm:  [ T. S. Ingebrigtsen, S. Toxvaerd, O. J. Heilmann, T. B. Schroeder and J. C. Dyre,
//                   ? ] 
// NVU validation: [ T. S. Ingebrigtsen, S. Toxvaerd, T. B. Schroeder and J. C. Dyre,
//                   ? ]
//////////////////////////////////////////////////////////////////////////////////////////////////////

template <class S> __global__ void integrateNVUAlgorithm( unsigned numParticles, float4* position, float4* velocity, 
							  float4* force, float4* image, S* simBox, float* simBoxPointer, 
							  float* constraintForce, float* displacementLengthCorrection, 
							  float* previousPotentialEnergy, float targetPotentialEnergy, 
							  float displacementLength ){
  
  if ( MyGP < numParticles ){ 
    float4 my_position = position[MyGP]; float4 my_velocity = velocity[MyGP]; 
    float4 my_force = force[MyGP]; float4 my_image = image[MyGP];
    
    float cForce = (constraintForce[0] + (previousPotentialEnergy[0] - targetPotentialEnergy)) / constraintForce[numParticles];
    float dispLengthCorrection = displacementLength / sqrtf(displacementLengthCorrection[0]);

    // Load the simulation box in 'local' memory to avoid bank conflicts.
    float array[simulationBoxSize];
    simBox->loadMemory( array, simBoxPointer );

    // Delta R(i+1/2).
    my_velocity.x = dispLengthCorrection * ( my_velocity.x + cForce * my_force.x );
    my_velocity.y = dispLengthCorrection * ( my_velocity.y + cForce * my_force.y ); 
    my_velocity.z = dispLengthCorrection * ( my_velocity.z + cForce * my_force.z );
    
    // R(i+1).
    my_position.x += my_velocity.x;
    my_position.y += my_velocity.y;
    my_position.z += my_velocity.z;

    float4 local_image = simBox->applyBoundaryCondition( my_position, array );
    
    // Save the integration in global memory.
    position[MyGP] = my_position; 
    velocity[MyGP] = my_velocity; 

    my_image.x += local_image.x;
    my_image.y += local_image.y;
    my_image.z += local_image.z;
    my_image.w += local_image.w;

    image[MyGP] = my_image;
  }
}

////////////////////////////////////////////////////////////
// Constraint Force/Numerical Correction 
////////////////////////////////////////////////////////////

template <class S> __global__ void calculateConstraintForce( unsigned numParticles, float4* position, float4* euclideanPosition,
							     float4* image, float4* euclideanImage, float4* velocity, float4* force, 
							     float* constraintForceArray, S* simBox, float* simBoxPointer ){
  if( MyGP < numParticles ){
    float4 my_position = position[MyGP]; float4 my_euclideanPosition = euclideanPosition[MyGP];
    float4 my_image = image[MyGP]; float4 my_euclideanImage = euclideanImage[MyGP];
    float4 my_velocity = velocity[MyGP]; float4 my_force = force[MyGP]; 
    
    // Load the simulation box in 'local' memory to avoid bank conflicts.
    float array[simulationBoxSize];
    simBox->loadMemory( array, simBoxPointer );

    float4 value = simBox->calculateDistanceWithImages( my_euclideanPosition, my_position, my_euclideanImage, my_image, array );

    constraintForceArray[MyGP] = -2.f * (my_velocity.x * my_force.x + my_velocity.y * my_force.y + my_velocity.z * my_force.z);
    constraintForceArray[numParticles+MyGP] = my_force.x * my_force.x + my_force.y * my_force.y + my_force.z * my_force.z;
    constraintForceArray[2*numParticles+MyGP] = value.w; 
  }
}

__global__ void calculateNumericalCorrection( unsigned numParticles, float4* velocity, float4* force, 
					      float* constraintForce, float* numericalCorrection, float* previousPotentialEnergy, 
					      float targetPotentialEnergy ){
  if( MyGP < numParticles ){
    float4 my_velocity = velocity[MyGP]; float4 my_force = force[MyGP]; 
    float constraintF = ( constraintForce[0] + (previousPotentialEnergy[0] - targetPotentialEnergy) ) / constraintForce[numParticles];
    float3 normVector = { my_velocity.x + constraintF * my_force.x, 
			  my_velocity.y + constraintF * my_force.y,  
			  my_velocity.z + constraintF * my_force.z };
    
    numericalCorrection[MyGP] = normVector.x * normVector.x + normVector.y * normVector.y + normVector.z * normVector.z;
    numericalCorrection[numParticles+MyGP] = my_force.w;
  }
}

////////////////////////////////////////////////////////////
// Zero the mean system velocity 
////////////////////////////////////////////////////////////

__global__ void zeroSystemVelocity( unsigned numParticles, float4* velocity, float4* particleVelocity ){
  if( MyGP < numParticles ){
    float4 my_velocity = velocity[MyGP]; 
    float4 totalVelocity = particleVelocity[0];
    float invTotalMass = 1.f / float(numParticles);
  
    // Subtract
    my_velocity.x -= invTotalMass * totalVelocity.x;
    my_velocity.y -= invTotalMass * totalVelocity.y;
    my_velocity.z -= invTotalMass * totalVelocity.z;
    velocity[MyGP] = my_velocity; 
  }
}

__global__ void calculateVelocity( unsigned numParticles, float4* velocity, float4* particleVelocity ){
  float4 my_velocity = { 0.f, 0.f, 0.f, 0.f };
  if( MyGP < numParticles ){
    particleVelocity[MyGP] = velocity[MyGP];
  }
  else{
    particleVelocity[MyGP] = my_velocity;
  }
}
