/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <iostream>
#include <iomanip>

#include "rumd/RUMD_Error.h"
#include "rumd/ParticleData.h"

using namespace std;

//////////////////////////////////////////////////
// CONSTRUCTION/DESTRUCTION 
//////////////////////////////////////////////////

ParticleData::ParticleData() {
  // These need to be zero before Allocate functions are first called.
  numParticles = 0;  
  numVirtualParticles = 0;
  allocatedNumParticles = 0;
  numBlocks = 0;
  verbose = true;
}

ParticleData::~ParticleData(){
  FreeParticles();
}

// Copies a small fraction of ParticleData.
ParticleData& ParticleData::operator=(const ParticleData& P){
  if(this != &P){ 
    cudaMemcpy( d_r, P.d_r, numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToDevice );
    cudaMemcpy( d_v, P.d_v, numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToDevice );
  }
  return *this;   
}


void ParticleData::CopyHostData(const ParticleData& P) {
  if(numParticles != P.GetNumberOfParticles())
    throw RUMD_Error("ParticleData","CopyHostData","Incompatible numParticles");
  // could call SetNumberOfParticles, but don't know pb and tp here!
  
  cudaMemcpy( h_r, P.h_r, numVirtualParticles * sizeof(float4), cudaMemcpyHostToHost );
  cudaMemcpy( h_v, P.h_v, numVirtualParticles * sizeof(float4), cudaMemcpyHostToHost );
  cudaMemcpy( h_f, P.h_f, numVirtualParticles * sizeof(float4), cudaMemcpyHostToHost );
  cudaMemcpy( h_w, P.h_w, numVirtualParticles * sizeof(float4), cudaMemcpyHostToHost );
  cudaMemcpy( h_im, P.h_im, numVirtualParticles * sizeof(float4), cudaMemcpyHostToHost );
  cudaMemcpy( h_sts, P.h_sts, numVirtualParticles * sizeof(float4), cudaMemcpyHostToHost );
  cudaMemcpy( h_misc, P.h_misc, numVirtualParticles * sizeof(float4), cudaMemcpyHostToHost );
  cudaMemcpy( h_Type, P.h_Type, numVirtualParticles * sizeof(unsigned int), cudaMemcpyHostToHost );
}

//////////////////////////////////////////////////
// MEMORY ALLOCATION/DEALLOCATION 
//////////////////////////////////////////////////

void ParticleData::SetNumberOfParticles(unsigned int set_numParticles, unsigned int pb, unsigned int tp){
  numParticles = set_numParticles;
  numBlocks = (numParticles + pb - 1) / pb;    
  numVirtualParticles = numBlocks * pb;

  if( !( (numParticles <= pb*numBlocks) && (numVirtualParticles >= numParticles) && (numVirtualParticles <  (numParticles + pb)) ) )
    throw(RUMD_Error("ParticleData","SetNumberOfParticles","ParticleData could not validate internal variables. Report this error to RUMD developers"));
  
  // Allocate memory.
  AllocateParticles(numVirtualParticles);
  
  // Print useful info.
  if(verbose) {
    cout << endl << "Setting number of particles and allocating on host and device:" << endl;
    cout << "Number of particles (N): " << setw( 9 ) << numParticles << endl;
    cout << "Particles per block (pb): " << setw( 6 ) <<  pb << endl;
    cout << "Threads per particle (tp): " << setw( 4 )<<  tp << endl;
    cout << "Number of blocks: " <<  setw( 14 ) << numBlocks << endl;
    cout << "Virtual particles: " <<  setw( 15 ) << numVirtualParticles << endl;
    cout << "Unused virtual particles: " << setw( 6 ) << (numVirtualParticles - numParticles) << endl;
    cout << "Threads per block: " << setw( 14 ) << (tp*pb) << endl;
    cout << "Number of threads: " <<  setw( 15 ) << (tp*pb*numBlocks) << endl << endl;
  }
}

void ParticleData::AllocateParticles(unsigned int nvp){
  if( allocatedNumParticles == nvp )
    return;
  
  FreeParticles();

  // Page-locked CPU Allocation
  if( cudaMallocHost( (void**) &h_r, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_r") );
  
  if( cudaMallocHost( (void**) &h_v, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_v") );
  
  if( cudaMallocHost( (void**) &h_f, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_f") );
  
  if( cudaMallocHost( (void**) &h_w, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_w") );
  
  if( cudaMallocHost( (void**) &h_im, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_im") );
  
  if( cudaMallocHost( (void**) &h_sts, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_sts") );
  
  if( cudaMallocHost( (void**) &h_misc, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_misc") );
  
  if( cudaMallocHost( (void**) &h_Type, nvp * sizeof(unsigned int) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_Type") );
  
  // GPU Allocation
  if( cudaMalloc( (void**) &d_r, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_r") );

  if( cudaMalloc( (void**) &d_v, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_v") );
  
  if( cudaMalloc( (void**) &d_f, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_f") );
  
  if( cudaMalloc( (void**) &d_w, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_w") );
  
  if( cudaMalloc( (void**) &d_im, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_im") );
  
  if( cudaMalloc( (void**) &d_sts, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_sts") );
  
  if( cudaMalloc( (void**) &d_misc, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_misc") );

  // Initialize all CPU memory to zero from the start.
  memset( h_r,      0, nvp * sizeof(float4) );
  memset( h_v,      0, nvp * sizeof(float4) );
  memset( h_f,      0, nvp * sizeof(float4) );
  memset( h_w,      0, nvp * sizeof(float4) );
  memset( h_im,     0, nvp * sizeof(float4) );
  memset( h_sts,    0, nvp * sizeof(float4) );
  memset( h_misc,   0, nvp * sizeof(float4) );
  memset( h_Type,   0, nvp * sizeof(unsigned int) );

  // Initialize all GPU memory to zero from the start (needed+safety).
  cudaMemset( d_r,      0, nvp * sizeof(float4) );
  cudaMemset( d_v,      0, nvp * sizeof(float4) );
  cudaMemset( d_f,      0, nvp * sizeof(float4) );
  cudaMemset( d_w,      0, nvp * sizeof(float4) );
  cudaMemset( d_im,     0, nvp * sizeof(float4) );
  cudaMemset( d_sts,    0, nvp * sizeof(float4) );
  cudaMemset( d_misc,   0, nvp * sizeof(float4) );
  
  if( cudaThreadSynchronize() != cudaSuccess )
    throw( RUMD_Error("ParticleData","AllocateParticles","Initialization failed on GPU") );
  
  allocatedNumParticles = nvp;
}

void ParticleData::FreeParticles() {
  if(allocatedNumParticles == 0)
    return;
  
  cudaFreeHost(h_r);
  cudaFreeHost(h_v);
  cudaFreeHost(h_f);
  cudaFreeHost(h_w);
  cudaFreeHost(h_im);
  cudaFreeHost(h_sts);
  cudaFreeHost(h_misc);
  cudaFreeHost(h_Type);
  cudaFree(d_r);
  cudaFree(d_v);
  cudaFree(d_f);
  cudaFree(d_w);
  cudaFree(d_im);
  cudaFree(d_sts);  
  cudaFree(d_misc);  
}

//////////////////////////////////////////////////
// Copy to the device.
//////////////////////////////////////////////////

void ParticleData::CopyPosToDevice() const{
  cudaMemcpy( d_r, h_r, numVirtualParticles * sizeof(float4), cudaMemcpyHostToDevice );
}

void ParticleData::CopyVelToDevice() const{
  cudaMemcpy( d_v, h_v, numVirtualParticles * sizeof(float4), cudaMemcpyHostToDevice );
}

void ParticleData::CopyForToDevice() const{
  cudaMemcpy( d_f, h_f, numVirtualParticles * sizeof(float4), cudaMemcpyHostToDevice );
}

void ParticleData::CopyImagesToDevice() const{
  cudaMemcpy( d_im, h_im, numVirtualParticles * sizeof(float4), cudaMemcpyHostToDevice );
}

void ParticleData::CopyConfToDevice() const{
  CopyPosToDevice();
  CopyVelToDevice();
  CopyForToDevice();
  CopyImagesToDevice();
  
  // Check if the copy generated errors.
  if( cudaThreadSynchronize() != cudaSuccess ) 
    throw( RUMD_Error("ParticleData","CopyConfToDevice","cudaMemcpy failed: simulation state => device") );
}

//////////////////////////////////////////////////
// Copy from the device.
//////////////////////////////////////////////////

void ParticleData::CopyPosFromDevice() const{
  cudaMemcpy( h_r, d_r, numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToHost );
}

void ParticleData::CopyVelFromDevice() const{
  cudaMemcpy( h_v, d_v,  numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToHost );
}

void ParticleData::CopyForFromDevice() const{
  cudaMemcpy( h_f, d_f, numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToHost );
}

void ParticleData::CopyVirFromDevice() const{
  cudaMemcpy( h_w, d_w, numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToHost );
}

void ParticleData::CopyImagesFromDevice() const{
  cudaMemcpy( h_im, d_im, numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToHost );
}

void ParticleData::CopyStressFromDevice() const{
  cudaMemcpy( h_sts, d_sts, numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToHost );
}

void ParticleData::CopyMiscFromDevice() const{ 
  cudaMemcpy( h_misc, d_misc, numVirtualParticles * sizeof(float4), cudaMemcpyDeviceToHost );
}

void ParticleData::CopyConfFromDevice() const{
  CopyPosFromDevice();
  CopyVelFromDevice();
  CopyForFromDevice();
  CopyVirFromDevice();
  CopyImagesFromDevice();
  CopyStressFromDevice();
  CopyMiscFromDevice();

  // Check if the copy generated errors.
  if( cudaThreadSynchronize() != cudaSuccess ) 
    throw( RUMD_Error("ParticleData","CopyConfFromDevice","cudaMemcpy failed: simulation state => host") );
}
